
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <asm/irq.h>

#include <asm/segment.h>
#include <asm/hardware.h>

// #define DEBUG

/* Administratrivia */
#ifndef EXT_IRQ_EXAMPLE_MAJOR
#define EXT_IRQ_EXAMPLE_MAJOR 62
#endif

#define ADS_EXT_IRQ(x) (IRQ_BOARD_START + (x))

// this is the vector for the external irq
#define EXT_IRQ ADS_EXT_IRQ(7)

static struct fasync_struct *fasync;


static void ext_irq_example_irq_handler(int irq, void *dev_id, struct pt_regs *regs )
{
	if (fasync)
	  kill_fasync(&fasync, SIGIO, POLL_IN);
}

static int ext_irq_example_register_irq(void)
{
	if (request_irq(EXT_IRQ, ext_irq_example_irq_handler, 0, "EXT_IRQ_EXAMPLE", NULL)) { 
             printk(KERN_WARNING "ext_irq_example.o: Could not allocate IRQ %d.\n", EXT_IRQ);
	     return -EINVAL;
	}

#if 0
	//  unmask the IRQ since the modified unmask routine will not do so!
	ADS_INT_EN1 |= (1 << (EXT_IRQ - ADS_EXT_IRQ(0)));
#endif
	
	return 0;
}

static int ext_irq_example_fasync(int fd, struct file *file, int on)
{
	int retval;

#ifdef DEBUG
	printk(KERN_DEBUG "in ext_irq_example_fasync\n");
#endif

	retval = fasync_helper(fd, file, on, &fasync);
	if (retval < 0) {
                printk(KERN_DEBUG "fasync_helper failed: %d\n", retval);
		return retval;
	}

	return ext_irq_example_register_irq();
}

static int ext_irq_example_fasync_release(int fd, struct file *file, int on)
{
	int retval;

#ifdef DEBUG
	printk(KERN_DEBUG "in ext_irq_example_fasync_release\n");
#endif

	fasync = 0;

	retval = fasync_helper(fd, file, on, &fasync);
	if (retval < 0) {
                printk(KERN_DEBUG "fasync_helper failed: %d\n", retval);
		return retval;
	}

#ifdef DEBUG
	printk(KERN_DEBUG "ext_irq_example_release: freeing irq: %d\n", EXT_IRQ);
#endif

	free_irq(EXT_IRQ, NULL);

	return retval;
}

int ext_irq_example_release(struct inode *inode, struct file *file)
{
	if (fasync)
          return ext_irq_example_fasync_release(-1, file, 0);
	else
	  return 0;
}

static struct file_operations ext_irq_example_fops = 
{
	fasync:  ext_irq_example_fasync,
	release: ext_irq_example_release,
};

int __init init_ext_irq_example(void)
{
	if (register_chrdev(EXT_IRQ_EXAMPLE_MAJOR,"ext_irq_example",&ext_irq_example_fops)) {
		printk("ext_irq_example: unable to get major %d\n", EXT_IRQ_EXAMPLE_MAJOR);
		return -EIO;
	}
	else {
		printk("ADS ext_irq_example driver initialized\n");
#ifdef DEBUG
		printk("EXT_IRQ: %d, irqoffset: %d, ADS_INT_ST1: %p\n", EXT_IRQ, EXT_IRQ - ADS_EXT_IRQ(0), &ADS_INT_ST1);
#endif
	}

	return 0;
}


void __init unregister_ext_irq_example_driver(void)
{
	unregister_chrdev(EXT_IRQ_EXAMPLE_MAJOR,"ext_irq_example");
}

module_init(init_ext_irq_example);
module_exit(unregister_ext_irq_example_driver);
