#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>

#define EXT_IRQ_EXAMPLE_MAJOR 62

#define EXT_IRQ_DEV "/dev/ext_irq"

void ext_irq_handler(int sig)
{
  if (sig == SIGIO)
    printf("Got IRQ\n");
  else
    printf("Got sig: %d (should be %d)\n", sig, SIGIO);
}

void ext_irq_test_install_irq_handler(int fd_irq)
{
  int res;
  int oflags;
  struct sigaction act;

  res = fcntl(fd_irq, F_SETOWN, getpid());
  if (res == -1) {
    perror("fcntl F_SETOWN");
    exit(res);
  }

  oflags = fcntl(fd_irq, F_GETFL);
  if (oflags == -1) {
    perror("fcntl F_GETFL");
    exit(res);
  }

  res = fcntl(fd_irq, F_SETFL, oflags | O_ASYNC);
  if (res == -1) {
    perror("fcntl F_SETFL");
    exit(res);
  }

  act.sa_handler = ext_irq_handler;

  res = sigaction(SIGIO, &act, NULL);
  if (res < 0) {
    perror("sigaction");
    exit(res);
  }

}

int main(int argc, char *argv[])
{
  int fd_irq;

  fd_irq = open(EXT_IRQ_DEV, O_RDWR);
  if (fd_irq < 0) {
    printf("Unable to open %s.  Try 'mknod %s c %d 0'\n", EXT_IRQ_DEV, EXT_IRQ_DEV, EXT_IRQ_EXAMPLE_MAJOR);
    printf(" or 'insmod ext_irq_example'\n");
    exit(fd_irq);
  }

  ext_irq_test_install_irq_handler(fd_irq);

  printf("This program prints a message whenever there is an interrupt on\n");
  printf("%s.\n\n", EXT_IRQ_DEV);

  while (1) {
    sleep(1);
  }
}
