#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>

#include "gpio_example.h"

#define GPIO_IRQ "/dev/gpio24"
#define GPIO_IN  "/dev/gpio25"
#define GPIO_OUT "/dev/gpio26"

void gpio_handler(int sig)
{
  if (sig == SIGIO)
    printf("Got IRQ\n");
  else
    printf("Got sig: %d (should be %d)\n", sig, SIGIO);
}

void gpio_test_install_irq_handler(int fd_irq)
{
  int res;
  int oflags;
  struct sigaction act;

  res = fcntl(fd_irq, F_SETOWN, getpid());
  if (res == -1) {
    perror("fcntl F_SETOWN");
    exit(res);
  }

  oflags = fcntl(fd_irq, F_GETFL);
  if (oflags == -1) {
    perror("fcntl F_GETFL");
    exit(res);
  }

  res = fcntl(fd_irq, F_SETFL, oflags | O_ASYNC);
  if (res == -1) {
    perror("fcntl F_SETFL");
    exit(res);
  }

  act.sa_handler = gpio_handler;

  res = sigaction(SIGIO, &act, NULL);
  if (res < 0) {
    perror("sigaction");
    exit(res);
  }

}

int main(int argc, char *argv[])
{
  int fd_irq;
  int fd_in;
  int fd_out;

  char in_value, out_value = '0';

  int res;

  fd_irq = open(GPIO_IRQ, O_RDWR);
  if (fd_irq < 0) {
    perror(GPIO_IRQ);
    exit(fd_irq);
  }

  fd_in  = open(GPIO_IN,  O_RDWR);
  if (fd_in < 0) {
    perror(GPIO_IN);
    exit(fd_in);
  }

  fd_out = open(GPIO_OUT, O_WRONLY);
  if (fd_out < 0) {
    perror(GPIO_OUT);
    exit(fd_out);
  }

  // set the irq pin for input and not alternate function
  res = ioctl(fd_irq, GPIO_EXAMPLE_CLEAR_GAFR);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_CLEAR_GAFR");
    exit(-1);
  }
  res = ioctl(fd_irq, GPIO_EXAMPLE_GPDR_INPUT);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_GPDR_INPUT");
    exit(-1);
  }

  // set the input pin for input and not alternate function
  res = ioctl(fd_in, GPIO_EXAMPLE_CLEAR_GAFR);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_CLEAR_GAFR");
    exit(-1);
  }
  res = ioctl(fd_in, GPIO_EXAMPLE_GPDR_INPUT);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_GPDR_INPUT");
    exit(-1);
  }

  // set the output pin for output and not alternate function
  res = ioctl(fd_out, GPIO_EXAMPLE_CLEAR_GAFR);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_CLEAR_GAFR");
    exit(-1);
  }
  res = ioctl(fd_out, GPIO_EXAMPLE_GPDR_OUTPUT);
  if (res < 0) {
    perror(GPIO_IRQ " GPIO_EXAMPLE_GPDR_OUTPUT");
    exit(-1);
  }

  gpio_test_install_irq_handler(fd_irq);

  printf("This program will loop forever.  Every 2 seconds it reads\n");
  printf("from GPIO_IN and writes alternating 0/1 to GPIO_OUT.  It\n");
  printf("also prints a message whenever there is an interrupt on\n");
  printf("GPIO_IRQ.\n\n");

  printf("GPIO_IRQ: %s, GPIO_IN: %s, GPIO_OUT: %s\n\n", GPIO_IRQ, GPIO_IN, GPIO_OUT);

  while (1) {

    res = read(fd_in, &in_value, 1);
    if (res != 1) {
      perror("Error reading");
      exit(res);
    }
    printf("GPIO_IN: %c\n", in_value);

    write(fd_out, &out_value, 1);
    if (res != 1) {
      perror("Error writing");
      exit(res);
    }

    if (out_value == '1')
      out_value = '0';
    else
      out_value = '1';

    // sleep 2 seconds or until SIGIO happens!
    sleep(2);
  }
}
